=begin
b4e2o.rb v1.01 (work with Evernote2Onenote v1.3.6.241)

The MIT License (MIT)

Copyright (c) 2024 H.Ogi <azarasister@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

=end

class EnPrefs
  attr_reader :prefs
  def initialize(conf=nil)
    @prefs ={
      "insert_tags" => 0,
      "tag_prefix" => "#",
      "tag_suffix" => "",
      "notebook_section" => true,
      "tag_section" => false,
      "include_tags" => "",
      "align_width" => 50,
      "mark_codeblock" => "",
      "mark_notelink" => ""
    }
    if conf
      open(conf, "r") {|io|
        while line = io.gets
          line.chomp!
          pref = line.split(/=/, 2)
          if pref.length == 2
            pref[0].strip!
            case pref[0]
            when "insert_tags", "align_width"
              @prefs[pref[0]] = pref[1].to_i
            when "notebook_section", "tag_section"
              @prefs[pref[0]] = pref[1] == "true"
            else
              @prefs[pref[0]] = pref[1]
            end
          end
        end
      }
    else
      puts "config.txt is not found. Run with default config."
    end
  end
end

class TagParser
  def initialize(section_tag=nil, prefs)
    @section_tag = prefs.prefs["notebook_section"] ? section_tag : nil
    @preserve = prefs.prefs["tag_section"]
    @tag_prefix = prefs.prefs["tag_prefix"]
    @tag_suffix = prefs.prefs["tag_suffix"]
    @include_tags = prefs.prefs["include_tags"]
    @tag_content = [];
  end

  def convert(tag_str)
    tag = ""
    if /<tag>(.+)<\/tag>/ =~ tag_str
      @tag_content.push(@tag_prefix + $1 + @tag_suffix)
      tag = $1
    end

    inc_tags = @include_tags.split(/\s*,\s*/)
    return @preserve || inc_tags.include?(tag) ? tag_str : ""
  end

  def getSectionTag()
    return @section_tag ? "    <tag>#{@section_tag}</tag>" : ""
  end

  def getTagContent()
    tag_str = @tag_content.join(" ")
    tag_str.gsub!(/\s/, "&nbsp;")
    "<div>#{tag_str}</div>"
  end
end

class ContentParser
  def initialize(prefs)
    @insert_tag_mode = prefs.prefs["insert_tags"] #0: none, 1: header, 2: footer
    @align_width = prefs.prefs["align_width"] * 2
    @mark_codeblock = prefs.prefs["mark_codeblock"]
    @mark_notelink = prefs.prefs["mark_notelink"]
  end

  def fixEmptyLine(line)
    line.gsub(/<div><br\/><\/div>/, "<div>&nbsp;<br/></div>")
  end

  def convertIndent(line)
    line.gsub(/(<div style="[^"]*padding-left:)(\d+)(px;[^"]*">)/) {
       indent = "&nbsp;" * ($2.to_i / 10)
      "#{$1}#{$2}#{$3}#{indent}"
    }
  end

  def convertCheckBox(line)
    line.gsub!(/<en-todo\/>/,"[ ]")
    line.gsub(/<en-todo checked="(\w+)" \/>/) {
      $1 == "true" ? "[x] " : "[ ]"
    }
  end

  def insertTags(line, tags)
    if @insert_tag_mode == 1
      line.sub(/<en-note[^>]*>/, "<en-note>#{tags}<div><br/><br/></div>")
    elsif @insert_tag_mode == 2
      line.sub(/<\/en-note>/, "<div><br/><br/></div>#{tags}</en-note>")
    else
      line
    end
  end

  def preprocessConvertingTask(line)
    line.gsub(/<div style="--en-task-group:true; --en-id:([^;]+);[^>]+>\s*<div [^>]+>[^<]+<\/div>\s*<div [^>]+>[^<]+<\/div>\s*<\/div>/) {
      "<!-- Task-Id:#{$1} -->"
    }
  end

  def convertTask(line, tasks)
    line.gsub(/<!-- Task-Id:(\S+) -->/) {
      task_array = tasks[$1]
      task_line = "<div>"
      task_array.each{|x|
        task_line += x.getTaskStr()
      }
      task_line
    }
  end

  def convertTextAlign(line)
    return line if @align_width < 1

    line.gsub!(/(<div style="[^"]*text-align:)(center|right)(;[^"]*">)(.+?)(<\/div>)/) {
      d1 = $1
      d2 = $2
      d3 = $3
      d4 = $4
      d5 = $5
      pre = $`

      if /<td[^>]*>\s*$/ =~ pre #exclude align in table
        "#{d1}#{d2}#{d3}#{d4}#{d5}"
      else
        aligned = d4.gsub(/&nbsp;/," ")
        aligned_content = aligned.gsub(/<[^>]+>/, "") #remove any tags inside div to align
        org_content = aligned_content
        if d2 == "center"
          aligned_content = aligned_content.center(@align_width)
        elsif d2 == "right"
          aligned_content = aligned_content.rjust(@align_width)
        end
        aligned_content.gsub!(/\s/, "&nbsp;")
        aligned.sub!(org_content, aligned_content)
        
        "#{d1}#{d2}#{d3}#{aligned}#{d5}"
      end
    }

    line
  end

  def convertCheckList(line)
    while line.index('<ul style="--en-todo:true;">')
      line = convertCheckListBlock(line)
    end

    line
  end

  def convertCheckListBlock(line)
    #scan start and last indexes of checklist considering nest
    start = line.index('<ul style="--en-todo:true;">')
    last = line.length - 1
    return nil if start == nil

    line.sub!(/<ul style="--en-todo:true;">/, "")

    nestLv = 0
    replaceFlg = [true]

    while nestLv >= 0
      ul_end = line.index('</ul>', start)
      ul_child = line.index(/<ul[^>]*>/, start) #find nested ul
      ul_todo_child = line.index('<ul style="--en-todo:true;">', start) #find nested en-todo ul
      if !ul_child || (ul_end && ul_end < ul_child) #end of en-todo
        last = ul_end
        line[last, 5] = "" #remove </ul> of en-todo
        line = convertCheckListMain(line, start, last, nestLv) if replaceFlg[nestLv]
        nestLv -= 1
      elsif ul_child && ul_child == ul_todo_child #nested en-todo
        line.sub!(/<ul style="--en-todo:true;">/, "")
        last = ul_child
        line = convertCheckListMain(line, start, last, nestLv) if replaceFlg[nestLv]
        nestLv += 1
        replaceFlg[nestLv] = true
      elsif ul_child && ul_child != ul_todo_child #nested ul but not en-todo
        last = ul_child
        line = convertCheckListMain(line, start, last, nestLv) if replaceFlg[nestLv]
        nestLv += 1
        replaceFlg[nestLv] = false
      else #other case
      end
    end

    line
  end

  def convertCheckListMain(line, start, last, nestLv)
    #convert
    substr = line.slice(start...last) #from start to before last
    indent = "&nbsp;" * (nestLv * 4)
    substr.gsub!(/<li style="--en-checked:false;">\s*<div>/, "<div>#{indent}[ ]")
    substr.gsub!(/<li style="--en-checked:true;">\s*<div>/, "<div>#{indent}[x]")
    substr.gsub!(/<li>\s*<div>/, "<div>#{indent}[ ]")
    substr.gsub!(/<\/li>/, '')

    #replace
    line[start...last] = substr
    line
  end

  def convertHR(line)
    line.gsub(/<hr\/>/,"------------------------------------------------------------------------------------")
  end

  def markCodeBlock(line)
    mark = @mark_codeblock
    if mark != ""
      pattern = /(<div style="--en-codeblock:true;)/
      line = markContent(pattern, mark, line)
    end
    line
  end

  def markNoteLink(line)
    mark = @mark_notelink
    if mark != ""
      pattern = /(<a [^>]*href="evernote:\/\/\/)/
      line = markContent(pattern, mark, line)
    end
    line
  end

  def markContent(pattern, mark, line)
    mark = mark.gsub(/</, "&lt;")
    mark = mark.gsub(/>/, "&gt;")
    mark = mark.gsub(/\s/, "&nbsp;")
    line.gsub(pattern)  {
      "<div>#{mark}</div>#{$1}"
    }
  end
end

class EnTask
  def initialize(io)
    @io = io
    @title = ""
    @status = "open" # open or completed
    @group_id = ""
  end

  def parse()
    while line = @io.gets
      /<([^\s>]+)>/ =~ line
      case ($1)
      when "title"
        /<title>([^<]+)<\/title>/ =~ line
        @title = $1
      when "taskGroupNoteLevelID"
        /<taskGroupNoteLevelID>([^<]+)<\/taskGroupNoteLevelID>/ =~ line
        @group_id = $1
      when "taskStatus"
        /<taskStatus>([^<]+)<\/taskStatus>/ =~ line
        @status = $1
      when "/task"
        @io = nil
        break
      else
      end
    end
  end

  def getGroupId()
    return @group_id
  end

  def getTaskStr()
    head = @status == "completed" ? "(x)" : "( )"
    return "<div>#{head} #{@title}</div>"
  end
end

def convertEnNote(line, content_parser, tag_parser)
  line = content_parser.fixEmptyLine(line)
  line = content_parser.convertIndent(line)
  line = content_parser.convertCheckBox(line)
  line = content_parser.convertHR(line)
  line = content_parser.convertTextAlign(line)
  tags = tag_parser.getTagContent()
  line = content_parser.insertTags(line, tags)
  line = content_parser.preprocessConvertingTask(line)
  line = content_parser.convertCheckList(line)
  line = content_parser.markCodeBlock(line)
  line = content_parser.markNoteLink(line)
end

def convert(file, prefs)
  section = File.basename(file, ".enex")
  section.strip! #creating section will fail if title includes space on head or tail
  tag_parser = TagParser.new(section, prefs)
  content_parser = ContentParser.new(prefs)
  tasks = {}
  tmp_file = open("#{section}.tmp", "w")
  open(file, "r") {|io|
    while line = io.gets
      line.chomp!
      /<([^>!?]+)>/ =~ line
      tag = $1 ? $1.split(/\s/)[0] : ""
      case (tag)
      when "title"
        /<title>(.+)<\/title>/ =~ line
        print "- Converting #{$1} ... "
        tmp_file.puts(line)
      when "tag"
        line = tag_parser.convert(line)
        tmp_file.puts(line) if line != ""
      when "en-note"
        new_line = line
        until (/<\/en-note>/ =~ line) #concat en-note lines
          line = io.gets
          line.chomp!
          new_line += line
        end
        new_line = convertEnNote(new_line, content_parser, tag_parser)
        tag_parser = TagParser.new(section, prefs)
        tmp_file.puts(new_line)
      when "content"
        tmp_file.puts(tag_parser.getSectionTag())
        tmp_file.puts(line)
      when "task"
        task = EnTask.new(io)
        task.parse()
        gid = task.getGroupId()
        if tasks[gid]
          tasks[gid].push(task)
        else
          tasks[gid] = [task];
        end
        # do not output line to remove task placeholder
      when "/note"
        print "done.\n"
        tmp_file.puts(line)
      else
        tmp_file.puts(line)
      end
    end
  }
  tmp_file.close()

  File.rename(file, "#{file}.old")

  if tasks.empty?
    File.rename("#{section}.tmp", "#{section}.enex")
  else
    puts
    new_file = open("#{section}.enex", "w")
    file = "#{section}.tmp"
    open(file, "r") {|io|
      while line = io.gets
        line.chomp!
        /<([^\s>]+)>/ =~ line
        case ($1)
        when "title"
          /<title>(.+)<\/title>/ =~ line
          print "- Postprocessing #{$1} ... "
          new_file.puts(line)
        when "en-note"
          line = content_parser.convertTask(line, tasks)
          new_file.puts(line)
        when "/note"
          print "done.\n"
          new_file.puts(line)
        else
          new_file.puts(line)
        end
      end
    }
    new_file.close()
    File.delete(file)
  end
end

def main
  puts "=== b4e2o v1.01 for Evernote2Onenote v1.3.6.241 ==="
  puts
  config_file = "config.txt"
  begin
    prefs = FileTest.exist?(config_file) ? EnPrefs.new(config_file) : EnPrefs.new()
    Dir.glob("*.enex").each {|file|
      unless FileTest.directory?(file)
        puts "[#{file}]"
        convert(file, prefs)
        puts
      end
    }
  rescue => ex
    puts
    p ex
    puts
  ensure
    print "Press Enter to exit: "
    str = gets
  end
end

main()


